@@ -10,11 +10,11 @@ void call_df();
 #define LOCK_FILE_NAME "egads.lock"
 #define PID_FILE_NAME  "egads.pid"
 #define SEED_FILE_NAME "egads.seed"
-#define EGADS_VERSION  "0.7"
-#define EGADS_DATE     "Feb 22, 2001"
+#define EGADS_VERSION  "0.9.5"
+#define EGADS_DATE     "September 2, 2002"
 
 #ifndef WIN32
-#define NUM_SOURCES    6 
+#define NUM_SOURCES    7 
 #else
 #define NUM_SOURCES    3
 #endif
@@ -27,6 +27,7 @@ void call_df();
 #define SRC_LOGFILE    3
 #define SRC_CMDS       4
 #define SRC_DEVRANDOM  5
+#define SRC_EXTERNAL   6
 #else
 #define SRC_PDH        2
 #endif
