@@ -75,6 +75,7 @@ static char *lock_file_name = EGADSDATA "/" LOCK_FILE_NAME;
 static char *pid_file_name = EGADSDATA "/" PID_FILE_NAME;
 static char *socket_file_name = EGADSDATA "/" SOCK_FILE_NAME;
 static char *seed_file_name = EGADSDATA "/" SEED_FILE_NAME;
+static char *egd_file_name = NULL;
 
 void timestamp(int sid)
 {
@@ -263,6 +264,46 @@ void *collect_entropy(void *arg)
   return NULL;
 }
 
+static
+int read_data(int fd, void *buffer, int len)
+{
+  int got, nb;
+
+  for (got = 0;  got < len;  got += nb)
+  {
+    nb = EGADS_read(fd, (char *)buffer + got, len - got);
+    if (nb == -1 && (errno == EAGAIN || errno == EINTR))
+    {
+      nb = 0;
+      continue;
+    }
+    if (nb <= 0)
+      return 0;
+  }
+
+  return 1;
+}
+
+static
+int write_data(int fd, void *buffer, int len)
+{
+  int done, nb;
+
+  for (done = 0;  done < len;  done += nb)
+  {
+    nb = EGADS_write(fd, (char *)buffer + done, len - done);
+    if (nb == -1 && (errno == EAGAIN || errno == EINTR))
+    {
+      nb = 0;
+      continue;
+    }
+    if (nb <= 0)
+      return 0;
+  }
+  
+  return 1;
+}
+
 /* Protocol is as follows:
  * 1 byte COMMAND, which must always be CMD_REQ_ENTROPY for right now.
  * arguments, command specific.
@@ -271,65 +312,163 @@ void *collect_entropy(void *arg)
  */
 
 static
-int process_entropy_request(int fd)
+int process_egads_request(int fd, char cmd)
 {
-  int howmuch;
+  int howmuch, rc;
   char *buffer;
 
-  if (!EGADS_read(fd, &howmuch, sizeof(int)))
+  switch (cmd)
   {
-    return 0;
+    case ECMD_REQ_ENTROPY:
+      if (!read_data(fd, &howmuch, sizeof(int)))
+      {
+        break;
+      }
+
+      EGADS_ALLOC(buffer, howmuch, 0);
+      EG_output(buffer, howmuch, 1);
+      rc = write_data(fd, buffer, howmuch);
+      EGADS_FREE(buffer);
+      if (!rc)
+      {
+        break;
+      }
+      EGADS_FREE(buffer);
+      return 0;
   }
 
-  EGADS_ALLOC(buffer, howmuch, 0);
-  EG_output(buffer, howmuch);
-
-  howmuch = EGADS_write(fd, buffer, howmuch);
-  EGADS_FREE(buffer);
-
-  return howmuch;
+  return 1;
 }
 
 static
-void *handle_client(void *arg)
+void *handle_egads_client(void *arg)
 {
-  int fd = (int)arg, nb;
+  int done = 0, fd = (int)arg;
   char cmd;
 
   pthread_detach(pthread_self());
-  for (;;)
+  while (!done)
   {
-    if (!(nb = EGADS_read(fd, &cmd, 1)))
+    if (!read_data(fd, &cmd, 1))
     {
-      close(fd);
-      return NULL;
+      done = 1;
     }
-
-    switch (cmd)
+    else
     {
-      case ECMD_REQ_ENTROPY:
-        if (!process_entropy_request(fd))
-        {
-          break;
-        }
-        // FALL THROUGH TO DEFAULT CASE
-
-      default:
-        close(fd);
-        return NULL;
+      done = process_egads_request(fd, cmd);
     }
   }
 
+  close(fd);
   return NULL;
 }
 
+static
+int process_egd_request(int fd, char cmd)
+{
+  int entropy;
+  char buffer[256];
+  unsigned char howmuch;
+
+  switch (cmd)
+  {
+    case EGD_REQ_ENTROPY_LEVEL:
+      entropy = (int)EG_entropy_level();
+      if (!write_data(fd, &entropy, sizeof(entropy)))
+      {
+        break;
+      }
+      return 0;
+
+    case EGD_REQ_ENTROPY_NB:
+      if (!read_data(fd, &howmuch, 1))
+      {
+        break;
+      }
+      buffer[0] = (unsigned char)(EG_output(&(buffer[1]), howmuch, 0) & 0xff);
+      if (!write_data(fd, buffer, (int)howmuch + 1))
+      {
+        break;
+      }
+      return 0;
+
+    case EGD_REQ_ENTROPY:
+      if (!read_data(fd, &howmuch, 1))
+      {
+        break;
+      }
+      if (!write_data(fd, buffer, EG_output(buffer, howmuch, 1)))
+      {
+        break;
+      }
+      return 0;
+
+    case EGD_ADD_ENTROPY:
+      if (!read_data(fd, buffer, 3))
+      {
+        break;
+      }
+      entropy = ((unsigned char)buffer[0] << 8) | (unsigned char)buffer[1];
+      howmuch = (unsigned char)buffer[2];
+      if (!read_data(fd, buffer, (int)howmuch))
+      {
+        break;
+      }
+      EG_add_entropy(SRC_EXTERNAL, buffer, howmuch, entropy);
+      return 0;
+
+    case EGD_REQ_PID:
+      sprintf(&(buffer[1]), "%d", getpid());
+      buffer[0] = (char)(strlen(buffer) & 0xff);
+      if (!write_data(fd, buffer, buffer[0] + 1))
+      {
+        break;
+      }
+      return 0;
+  }
+
+  return 1;
+}
+
+static
+void *handle_egd_client(void *arg)
+{
+  int fd = (int)arg, done = 0;
+  char cmd;
+
+  pthread_detach(pthread_self());
+  while (!done)
+  {
+    if (!read_data(fd, &cmd, 1))
+    {
+      done = 1;
+    }
+    else
+    {
+      done = process_egd_request(fd, cmd);
+    }
+  }
+
+  close(fd);
+  return NULL;
+}
+
+struct server_arg
+{
+  int sfd;
+  int egd;
+};
+
 static
 void *server_main(void *arg)
 {
-  int cfd, len, sfd = (int)arg;
+  int cfd, len, egd, sfd, old;
   pthread_t tid;
   struct sockaddr_un csa;
-  int	 old;
+
+  sfd = ((struct server_arg *)arg)->sfd;
+  egd = ((struct server_arg *)arg)->egd;
+  EGADS_FREE(arg);
 
   pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &old);
   pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &old);
@@ -346,16 +485,24 @@ void *server_main(void *arg)
     }
 
     /* TODO: maybe limit the number of simultaneous clients? */
-    pthread_create(&tid, NULL, handle_client, (void *)cfd);
+    if (!egd)
+    {
+      pthread_create(&tid, NULL, handle_egads_client, (void *)cfd);
+    }
+    else
+    {
+      pthread_create(&tid, NULL, handle_egd_client, (void *)cfd);
+    }
   }
 }
 
 static
-pthread_t run_server(char *file)
+pthread_t run_server(char *file, int egd)
 {
   int rc, sfd;
   pthread_t tid;
   struct sockaddr_un ssa;
+  struct server_arg *info;
 
   ssa.sun_family = AF_UNIX;
   strncpy(ssa.sun_path, file, sizeof(ssa.sun_path) - 1);
@@ -368,6 +515,7 @@ pthread_t run_server(char *file)
     exit(rc);
   }
 
+  umask(0);
   if (bind(sfd, (struct sockaddr *)&ssa, sizeof(ssa)) == -1)
   {
     if (errno == EADDRINUSE)
@@ -390,6 +538,7 @@ pthread_t run_server(char *file)
       exit(rc);
     }
   }
+  umask(066);
 
   if (listen(sfd, SOMAXCONN) == -1)
   {
@@ -398,7 +547,13 @@ pthread_t run_server(char *file)
     exit(rc);
   }
 
-  pthread_create(&tid, NULL, server_main, (void *)sfd);
+  EGADS_ALLOC(info, sizeof(info), 0);
+  info->sfd = sfd;
+  info->egd = egd;
+  if (pthread_create(&tid, NULL, server_main, info))
+  {
+    EGADS_FREE(info);
+  }
   return tid;
 }
 
@@ -503,6 +658,7 @@ void display_help(char *progname)
 {
   fprintf(stderr, "usage: %s [dhlpvCFLRSTV]\n\n", progname);
   fprintf(stderr, "-d <seconds>  Specify the delay between collections\n");
+  fprintf(stderr, "-e <name>     Specify the name of a socket to use for EGD\n");
   fprintf(stderr, "-h            Display this list of options\n");
   fprintf(stderr, "-l <logfile>  Specify a log file to watch\n");
   fprintf(stderr, "-p <path>     Specify the data directory to use\n");
@@ -521,7 +677,7 @@ void read_options(int argc, char **argv)
 {
   int i;
 
-  while ((i = getopt(argc, argv, "d:hl:p:vCFLRSTV?")) != -1)
+  while ((i = getopt(argc, argv, "d:e:hl:p:vCFLRSTV?")) != -1)
   {
     switch (i)
     {
@@ -533,6 +689,14 @@ void read_options(int argc, char **argv)
         }
         break;
 
+      case 'e':
+        if (egd_file_name)
+        {
+          EGADS_FREE(egd_file_name);
+        }
+        egd_file_name = EGADS_STRDUP(optarg);
+        break;
+
       case 'l':
         add_logfile(EGADS_STRDUP(optarg));
         break;
@@ -769,8 +933,8 @@ void save_me(void)
 int main(int argc, char **argv)
 {
   int threadc, which;
-  sigset_t mask;
-  pthread_t tid, *threadv;
+  sigset_t mask, oldmask;
+  pthread_t egdtid, tid, *threadv;
 
   umask(066);
   read_options(argc, argv);
@@ -784,7 +948,19 @@ int main(int argc, char **argv)
   write_pid();
   open_log_files();
   threadv = ACCUM_init(&threadc);
-  tid = run_server(socket_file_name);
+
+  tid = run_server(socket_file_name, 0);
+  if (egd_file_name)
+  {
+    egdtid = run_server(egd_file_name, 1);
+  }
+  else
+  {
+    /* cast here to avoid compiler warnings.  pthread_t is opaque so it
+     * could be either an int or a pointer.
+     */
+    egdtid = (pthread_t)NULL;
+  }
 
   sigemptyset(&mask);
   sigaddset(&mask, SIGHUP);
@@ -792,13 +968,29 @@ int main(int argc, char **argv)
   sigaddset(&mask, SIGINT);
   sigaddset(&mask, SIGQUIT);
   sigaddset(&mask, SIGALRM);
-  pthread_sigmask(SIG_BLOCK, &mask, NULL);
+#if !defined(__APPLE__) || !defined(__MACH__)
+  pthread_sigmask(SIG_BLOCK, &mask, &oldmask);
+#else
+  sigprocmask(SIG_BLOCK, &mask, &oldmask);
+#endif
 
   do
   {
+#if defined(__APPLE__) && defined(__MACH__)
+    sigset_t  pending;
+
+    sigsuspend(&oldmask);
+    sigpending(&pending);
+    if (!sigismember(&pending, SIGALRM))
+      which = !SIGALRM;   /* who cares */
+    else
+    {
+      which = SIGALRM;
+#else
     sigwait(&mask, &which);
     if (which == SIGALRM)
     {
+#endif
       if (truerand_count)
       {
         truerand_done = 1;
@@ -811,6 +1003,12 @@ int main(int argc, char **argv)
   }
   while (which == SIGALRM);
 
+  if (egd_file_name)
+  {
+    pthread_cancel(egdtid);
+    pthread_join(egdtid, NULL);
+  }
+
   pthread_cancel(tid);
   pthread_join(tid, NULL);
 
